@@ -1,38 +1,55 @@
 """Type checking of attribute access"""
 
-from typing import cast, Callable, List, Optional
+from typing import cast, Callable, List, Optional, TypeVar
 
 from mypy.types import (
     Type, Instance, AnyType, TupleType, CallableType, FunctionLike, TypeVarDef,
-    Overloaded, TypeVarType, TypeTranslator, UnionType, PartialType,
-    DeletedType, NoneTyp, TypeType
+    Overloaded, TypeVarType, UnionType, PartialType,
+    DeletedType, NoneTyp, TypeType, function_type
 )
-from mypy.nodes import TypeInfo, FuncBase, Var, FuncDef, SymbolNode, Context
-from mypy.nodes import ARG_POS, ARG_STAR, ARG_STAR2, function_type, Decorator, OverloadedFuncDef
+from mypy.nodes import TypeInfo, FuncBase, Var, FuncDef, SymbolNode, Context, MypyFile
+from mypy.nodes import ARG_POS, ARG_STAR, ARG_STAR2
+from mypy.nodes import Decorator, OverloadedFuncDef
 from mypy.messages import MessageBuilder
 from mypy.maptype import map_instance_to_supertype
-from mypy.expandtype import expand_type_by_instance
-from mypy.nodes import method_type, method_type_with_fallback
-from mypy.semanal import self_type
+from mypy.expandtype import expand_type_by_instance, expand_type
+from mypy.infer import infer_type_arguments
+from mypy.semanal import fill_typevars
 from mypy import messages
 from mypy import subtypes
+MYPY = False
+if MYPY:  # import for forward declaration only
+    import mypy.checker
+
+from mypy import experiments
 
 
-def analyze_member_access(name: str, typ: Type, node: Context, is_lvalue: bool,
+def analyze_member_access(name: str,
+                          typ: Type,
+                          node: Context,
+                          is_lvalue: bool,
                           is_super: bool,
+                          is_operator: bool,
                           builtin_type: Callable[[str], Instance],
                           not_ready_callback: Callable[[str, Context], None],
-                          msg: MessageBuilder, override_info: TypeInfo = None,
-                          report_type: Type = None) -> Type:
-    """Analyse attribute access.
+                          msg: MessageBuilder, *,
+                          original_type: Type,
+                          override_info: TypeInfo = None,
+                          chk: 'mypy.checker.TypeChecker' = None) -> Type:
+    """Return the type of attribute `name` of typ.
 
     This is a general operation that supports various different variations:
 
       1. lvalue or non-lvalue access (i.e. setter or getter access)
       2. supertype access (when using super(); is_super == True and
          override_info should refer to the supertype)
+
+    original_type is the most precise inferred or declared type of the base object
+    that we have available. typ is generally a supertype of original_type.
+    When looking for an attribute of typ, we may perform recursive calls targeting
+    the fallback type, for example.
+    original_type is always the type used in the initial call.
     """
-    report_type = report_type or typ
     if isinstance(typ, Instance):
         if name == '__init__' and not is_super:
             # Accessing __init__ in statically typed code would compromise
@@ -46,45 +63,59 @@ def analyze_member_access(name: str, typ: Type, node: Context, is_lvalue: bool,
         if override_info:
             info = override_info
 
+        if (experiments.find_occurrences and
+                info.name() == experiments.find_occurrences[0] and
+                name == experiments.find_occurrences[1]):
+            msg.note("Occurrence of '{}.{}'".format(*experiments.find_occurrences), node)
+
         # Look up the member. First look up the method dictionary.
         method = info.get_method(name)
         if method:
             if method.is_property:
                 assert isinstance(method, OverloadedFuncDef)
-                method = cast(OverloadedFuncDef, method)
                 return analyze_var(name, method.items[0].var, typ, info, node, is_lvalue, msg,
-                                   not_ready_callback)
+                                   original_type, not_ready_callback)
             if is_lvalue:
                 msg.cant_assign_to_method(node)
-            typ = map_instance_to_supertype(typ, method.info)
+            signature = function_type(method, builtin_type('builtins.function'))
             if name == '__new__':
                 # __new__ is special and behaves like a static method -- don't strip
                 # the first argument.
-                signature = function_type(method, builtin_type('builtins.function'))
+                pass
             else:
-                signature = method_type_with_fallback(method, builtin_type('builtins.function'))
+                signature = bind_self(signature, original_type)
+            typ = map_instance_to_supertype(typ, method.info)
             return expand_type_by_instance(signature, typ)
         else:
             # Not a method.
             return analyze_member_var_access(name, typ, info, node,
                                              is_lvalue, is_super, builtin_type,
                                              not_ready_callback, msg,
-                                             report_type=report_type)
+                                             original_type=original_type, chk=chk)
     elif isinstance(typ, AnyType):
         # The base object has dynamic type.
         return AnyType()
+    elif isinstance(typ, NoneTyp):
+        if chk and chk.should_suppress_optional_error([typ]):
+            return AnyType()
+        # The only attribute NoneType has are those it inherits from object
+        return analyze_member_access(name, builtin_type('builtins.object'), node, is_lvalue,
+                                     is_super, is_operator, builtin_type, not_ready_callback, msg,
+                                     original_type=original_type, chk=chk)
     elif isinstance(typ, UnionType):
         # The base object has dynamic type.
         msg.disable_type_names += 1
-        results = [analyze_member_access(name, subtype, node, is_lvalue,
-                                         is_super, builtin_type, not_ready_callback, msg)
+        results = [analyze_member_access(name, subtype, node, is_lvalue, is_super,
+                                         is_operator, builtin_type, not_ready_callback, msg,
+                                         original_type=original_type, chk=chk)
                    for subtype in typ.items]
         msg.disable_type_names -= 1
         return UnionType.make_simplified_union(results)
     elif isinstance(typ, TupleType):
         # Actually look up from the fallback instance type.
-        return analyze_member_access(name, typ.fallback, node, is_lvalue,
-                                     is_super, builtin_type, not_ready_callback, msg)
+        return analyze_member_access(name, typ.fallback, node, is_lvalue, is_super,
+                                     is_operator, builtin_type, not_ready_callback, msg,
+                                     original_type=original_type, chk=chk)
     elif isinstance(typ, FunctionLike) and typ.is_type_obj():
         # Class attribute.
         # TODO super?
@@ -92,25 +123,40 @@ def analyze_member_access(name: str, typ: Type, node: Context, is_lvalue: bool,
         if isinstance(ret_type, TupleType):
             ret_type = ret_type.fallback
         if isinstance(ret_type, Instance):
-            result = analyze_class_attribute_access(ret_type, name, node, is_lvalue,
-                                                    builtin_type, not_ready_callback, msg)
-            if result:
-                return result
+            if not is_operator:
+                # When Python sees an operator (eg `3 == 4`), it automatically translates that
+                # into something like `int.__eq__(3, 4)` instead of `(3).__eq__(4)` as an
+                # optimation.
+                #
+                # While it normally it doesn't matter which of the two versions are used, it
+                # does cause inconsistencies when working with classes. For example, translating
+                # `int == int` to `int.__eq__(int)` would not work since `int.__eq__` is meant to
+                # compare two int _instances_. What we really want is `type(int).__eq__`, which
+                # is meant to compare two types or classes.
+                #
+                # This check makes sure that when we encounter an operator, we skip looking up
+                # the corresponding method in the current instance to avoid this edge case.
+                # See https://github.com/python/mypy/pull/1787 for more info.
+                result = analyze_class_attribute_access(ret_type, name, node, is_lvalue,
+                                                        builtin_type, not_ready_callback, msg,
+                                                        original_type=original_type)
+                if result:
+                    return result
             # Look up from the 'type' type.
             return analyze_member_access(name, typ.fallback, node, is_lvalue, is_super,
-                                         builtin_type, not_ready_callback, msg,
-                                         report_type=report_type)
+                                         is_operator, builtin_type, not_ready_callback, msg,
+                                         original_type=original_type, chk=chk)
         else:
             assert False, 'Unexpected type {}'.format(repr(ret_type))
     elif isinstance(typ, FunctionLike):
         # Look up from the 'function' type.
         return analyze_member_access(name, typ.fallback, node, is_lvalue, is_super,
-                                     builtin_type, not_ready_callback, msg,
-                                     report_type=report_type)
+                                     is_operator, builtin_type, not_ready_callback, msg,
+                                     original_type=original_type, chk=chk)
     elif isinstance(typ, TypeVarType):
         return analyze_member_access(name, typ.upper_bound, node, is_lvalue, is_super,
-                                     builtin_type, not_ready_callback, msg,
-                                     report_type=report_type)
+                                     is_operator, builtin_type, not_ready_callback, msg,
+                                     original_type=original_type, chk=chk)
     elif isinstance(typ, DeletedType):
         msg.deleted_as_rvalue(typ, node)
         return AnyType()
@@ -122,16 +168,21 @@ def analyze_member_access(name: str, typ: Type, node: Context, is_lvalue: bool,
         elif isinstance(typ.item, TypeVarType):
             if isinstance(typ.item.upper_bound, Instance):
                 item = typ.item.upper_bound
-        if item:
+        if item and not is_operator:
+            # See comment above for why operators are skipped
             result = analyze_class_attribute_access(item, name, node, is_lvalue,
-                                                    builtin_type, not_ready_callback, msg)
+                                                    builtin_type, not_ready_callback, msg,
+                                                    original_type=original_type)
             if result:
                 return result
         fallback = builtin_type('builtins.type')
         return analyze_member_access(name, fallback, node, is_lvalue, is_super,
-                                     builtin_type, not_ready_callback, msg,
-                                     report_type=report_type)
-    return msg.has_no_attr(report_type, name, node)
+                                     is_operator, builtin_type, not_ready_callback, msg,
+                                     original_type=original_type, chk=chk)
+
+    if chk and chk.should_suppress_optional_error([typ]):
+        return AnyType()
+    return msg.has_no_attr(original_type, name, node)
 
 
 def analyze_member_var_access(name: str, itype: Instance, info: TypeInfo,
@@ -139,11 +190,13 @@ def analyze_member_var_access(name: str, itype: Instance, info: TypeInfo,
                               builtin_type: Callable[[str], Instance],
                               not_ready_callback: Callable[[str, Context], None],
                               msg: MessageBuilder,
-                              report_type: Type = None) -> Type:
+                              original_type: Type,
+                              chk: 'mypy.checker.TypeChecker' = None) -> Type:
     """Analyse attribute access that does not target a method.
 
-    This is logically part of analyze_member_access and the arguments are
-    similar.
+    This is logically part of analyze_member_access and the arguments are similar.
+
+    original_type is the type of E in the expression E.var
     """
     # It was not a method. Try looking up a variable.
     v = lookup_member_var_or_accessor(info, name, is_lvalue)
@@ -152,18 +205,19 @@ def analyze_member_var_access(name: str, itype: Instance, info: TypeInfo,
     if isinstance(vv, Decorator):
         # The associated Var node of a decorator contains the type.
         v = vv.var
-
     if isinstance(v, Var):
-        return analyze_var(name, v, itype, info, node, is_lvalue, msg, not_ready_callback)
+        return analyze_var(name, v, itype, info, node, is_lvalue, msg,
+                           original_type, not_ready_callback)
     elif isinstance(v, FuncDef):
         assert False, "Did not expect a function"
     elif not v and name not in ['__getattr__', '__setattr__']:
         if not is_lvalue:
             method = info.get_method('__getattr__')
             if method:
+                function = function_type(method, builtin_type('builtins.function'))
+                bound_method = bind_self(function, original_type)
                 typ = map_instance_to_supertype(itype, method.info)
-                getattr_type = expand_type_by_instance(
-                    method_type_with_fallback(method, builtin_type('builtins.function')), typ)
+                getattr_type = expand_type_by_instance(bound_method, typ)
                 if isinstance(getattr_type, CallableType):
                     return getattr_type.ret_type
 
@@ -175,15 +229,19 @@ def analyze_member_var_access(name: str, itype: Instance, info: TypeInfo,
         msg.undefined_in_superclass(name, node)
         return AnyType()
     else:
-        return msg.has_no_attr(report_type or itype, name, node)
+        if chk and chk.should_suppress_optional_error([itype]):
+            return AnyType()
+        return msg.has_no_attr(original_type, name, node)
 
 
 def analyze_var(name: str, var: Var, itype: Instance, info: TypeInfo, node: Context,
-               is_lvalue: bool, msg: MessageBuilder,
-               not_ready_callback: Callable[[str, Context], None]) -> Type:
+                is_lvalue: bool, msg: MessageBuilder, original_type: Type,
+                not_ready_callback: Callable[[str, Context], None]) -> Type:
     """Analyze access to an attribute via a Var node.
 
     This is conceptually part of analyze_member_access and the arguments are similar.
+
+    original_type is the type of E in the expression E.var
     """
     # Found a member variable.
     itype = map_instance_to_supertype(itype, var.info)
@@ -192,7 +250,10 @@ def analyze_var(name: str, var: Var, itype: Instance, info: TypeInfo, node: Cont
         if isinstance(typ, PartialType):
             return handle_partial_attribute_type(typ, is_lvalue, msg, var)
         t = expand_type_by_instance(typ, itype)
-        if var.is_initialized_in_class and isinstance(t, FunctionLike):
+        if is_lvalue and var.is_property and not var.is_settable_property:
+            # TODO allow setting attributes in subclass (although it is probably an error)
+            msg.read_only_property(name, info, node)
+        if var.is_initialized_in_class and isinstance(t, FunctionLike) and not t.is_type_obj():
             if is_lvalue:
                 if var.is_property:
                     if not var.is_settable_property:
@@ -204,13 +265,14 @@ def analyze_var(name: str, var: Var, itype: Instance, info: TypeInfo, node: Cont
                 # Class-level function objects and classmethods become bound
                 # methods: the former to the instance, the latter to the
                 # class.
-                functype = cast(FunctionLike, t)
+                functype = t
                 check_method_type(functype, itype, var.is_classmethod, node, msg)
-                signature = method_type(functype)
+                signature = bind_self(functype, original_type)
                 if var.is_property:
                     # A property cannot have an overloaded type => the cast
                     # is fine.
-                    return cast(CallableType, signature).ret_type
+                    assert isinstance(signature, CallableType)
+                    return signature.ret_type
                 else:
                     return signature
         return t
@@ -236,7 +298,7 @@ def handle_partial_attribute_type(typ: PartialType, is_lvalue: bool, msg: Messag
 
 
 def lookup_member_var_or_accessor(info: TypeInfo, name: str,
-                                  is_lvalue: bool) -> SymbolNode:
+                                  is_lvalue: bool) -> Optional[SymbolNode]:
     """Find the attribute/accessor node that refers to a member of a type."""
     # TODO handle lvalues
     node = info.get(name)
@@ -259,7 +321,7 @@ def check_method_type(functype: FunctionLike, itype: Instance, is_classmethod: b
             # we passed to typ.fallback in analyze_member_access. See #1432.
             if isinstance(selfarg, TupleType):
                 selfarg = selfarg.fallback
-            if not subtypes.is_equivalent(selfarg, itype):
+            if not subtypes.is_subtype(selfarg, itype):
                 msg.invalid_method_type(item, context)
         else:
             # Check that cls argument has type 'Any' or valid class type.
@@ -281,7 +343,9 @@ def analyze_class_attribute_access(itype: Instance,
                                    is_lvalue: bool,
                                    builtin_type: Callable[[str], Instance],
                                    not_ready_callback: Callable[[str, Context], None],
-                                   msg: MessageBuilder) -> Type:
+                                   msg: MessageBuilder,
+                                   original_type: Type) -> Type:
+    """original_type is the type of E in the expression E.var"""
     node = itype.type.get(name)
     if not node:
         if itype.type.fallback_to_any:
@@ -304,7 +368,7 @@ def analyze_class_attribute_access(itype: Instance,
         if isinstance(t, PartialType):
             return handle_partial_attribute_type(t, is_lvalue, msg, node.node)
         is_classmethod = is_decorated and cast(Decorator, node.node).func.is_class
-        return add_class_tvars(t, itype.type, is_classmethod, builtin_type)
+        return add_class_tvars(t, itype, is_classmethod, builtin_type, original_type)
     elif isinstance(node.node, Var):
         not_ready_callback(name, context)
         return AnyType()
@@ -312,6 +376,10 @@ def analyze_class_attribute_access(itype: Instance,
     if isinstance(node.node, TypeInfo):
         return type_object_type(node.node, builtin_type)
 
+    if isinstance(node.node, MypyFile):
+        # Reference to a module object.
+        return builtin_type('builtins.module')
+
     if is_decorated:
         # TODO: Return type of decorated function. This is quick hack to work around #998.
         return AnyType()
@@ -319,25 +387,35 @@ def analyze_class_attribute_access(itype: Instance,
         return function_type(cast(FuncBase, node.node), builtin_type('builtins.function'))
 
 
-def add_class_tvars(t: Type, info: TypeInfo, is_classmethod: bool,
-                    builtin_type: Callable[[str], Instance]) -> Type:
+def add_class_tvars(t: Type, itype: Instance, is_classmethod: bool,
+                    builtin_type: Callable[[str], Instance],
+                    original_type: Type) -> Type:
+    """Instantiate type variables during analyze_class_attribute_access,
+    e.g T and Q in the following:
+
+    def A(Generic(T)):
+        @classmethod
+        def foo(cls: Type[Q]) -> Tuple[T, Q]: ...
+
+    class B(A): pass
+
+    B.foo()
+
+    original_type is the value of the type B in the expression B.foo()
+    """
+    # TODO: verify consistency between Q and T
+    info = itype.type  # type: TypeInfo
     if isinstance(t, CallableType):
         # TODO: Should we propagate type variable values?
-        vars = [TypeVarDef(n, i + 1, None, builtin_type('builtins.object'), tv.variance)
-                for (i, n), tv in zip(enumerate(info.type_vars), info.defn.type_vars)]
-        arg_types = t.arg_types
-        arg_kinds = t.arg_kinds
-        arg_names = t.arg_names
+        tvars = [TypeVarDef(n, i + 1, None, builtin_type('builtins.object'), tv.variance)
+                 for (i, n), tv in zip(enumerate(info.type_vars), info.defn.type_vars)]
         if is_classmethod:
-            arg_types = arg_types[1:]
-            arg_kinds = arg_kinds[1:]
-            arg_names = arg_names[1:]
-        return t.copy_modified(arg_types=arg_types, arg_kinds=arg_kinds, arg_names=arg_names,
-                               variables=vars + t.variables)
+            t = bind_self(t, original_type)
+        return t.copy_modified(variables=tvars + t.variables)
     elif isinstance(t, Overloaded):
-        return Overloaded([cast(CallableType, add_class_tvars(i, info, is_classmethod,
-                                                              builtin_type))
-                           for i in t.items()])
+        return Overloaded([cast(CallableType, add_class_tvars(item, itype, is_classmethod,
+                                                              builtin_type, original_type))
+                           for item in t.items()])
     return t
 
 
@@ -380,7 +458,7 @@ def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) ->
 
 def type_object_type_from_function(init_or_new: FuncBase, info: TypeInfo,
                                    fallback: Instance) -> FunctionLike:
-    signature = method_type_with_fallback(init_or_new, fallback)
+    signature = bind_self(function_type(init_or_new, fallback))
 
     # The __init__ method might come from a generic superclass
     # (init_or_new.info) with type variables that do not map
@@ -404,8 +482,9 @@ def type_object_type_from_function(init_or_new: FuncBase, info: TypeInfo,
         return class_callable(signature, info, fallback, special_sig)
     else:
         # Overloaded __init__/__new__.
+        assert isinstance(signature, Overloaded)
         items = []  # type: List[CallableType]
-        for item in cast(Overloaded, signature).items():
+        for item in signature.items():
             items.append(class_callable(item, info, fallback, special_sig))
         return Overloaded(items)
 
@@ -414,51 +493,15 @@ def class_callable(init_type: CallableType, info: TypeInfo, type_type: Instance,
                    special_sig: Optional[str]) -> CallableType:
     """Create a type object type based on the signature of __init__."""
     variables = []  # type: List[TypeVarDef]
-    for i, tvar in enumerate(info.defn.type_vars):
-        variables.append(TypeVarDef(tvar.name, i + 1, tvar.values, tvar.upper_bound,
-                                    tvar.variance))
-
-    initvars = init_type.variables
-    variables.extend(initvars)
+    variables.extend(info.defn.type_vars)
+    variables.extend(init_type.variables)
 
     callable_type = init_type.copy_modified(
-        ret_type=self_type(info), fallback=type_type, name=None, variables=variables,
+        ret_type=fill_typevars(info), fallback=type_type, name=None, variables=variables,
         special_sig=special_sig)
     c = callable_type.with_name('"{}"'.format(info.name()))
-    cc = convert_class_tvars_to_func_tvars(c, len(initvars))
-    cc.is_classmethod_class = True
-    return cc
-
-
-def convert_class_tvars_to_func_tvars(callable: CallableType,
-                                      num_func_tvars: int) -> CallableType:
-    return cast(CallableType, callable.accept(TvarTranslator(num_func_tvars)))
-
-
-class TvarTranslator(TypeTranslator):
-    def __init__(self, num_func_tvars: int) -> None:
-        super().__init__()
-        self.num_func_tvars = num_func_tvars
-
-    def visit_type_var(self, t: TypeVarType) -> Type:
-        if t.id < 0:
-            return t
-        else:
-            return TypeVarType(t.name, -t.id - self.num_func_tvars, t.values, t.upper_bound,
-                               t.variance)
-
-    def translate_variables(self,
-                            variables: List[TypeVarDef]) -> List[TypeVarDef]:
-        if not variables:
-            return variables
-        items = []  # type: List[TypeVarDef]
-        for v in variables:
-            if v.id > 0:
-                items.append(TypeVarDef(v.name, -v.id - self.num_func_tvars,
-                                        v.values, v.upper_bound, v.variance))
-            else:
-                items.append(v)
-        return items
+    c.is_classmethod_class = True
+    return c
 
 
 def map_type_from_supertype(typ: Type, sub_info: TypeInfo,
@@ -475,7 +518,7 @@ def map_type_from_supertype(typ: Type, sub_info: TypeInfo,
     Now S in the context of D would be mapped to E[T] in the context of C.
     """
     # Create the type of self in subtype, of form t[a1, ...].
-    inst_type = self_type(sub_info)
+    inst_type = fill_typevars(sub_info)
     if isinstance(inst_type, TupleType):
         inst_type = inst_type.fallback
     # Map the type of self to supertype. This gets us a description of the
@@ -489,3 +532,82 @@ def map_type_from_supertype(typ: Type, sub_info: TypeInfo,
     # in inst_type they are interpreted in subtype context. This works even if
     # the names of type variables in supertype and subtype overlap.
     return expand_type_by_instance(typ, inst_type)
+
+
+F = TypeVar('F', bound=FunctionLike)
+
+
+def bind_self(method: F, original_type: Type = None) -> F:
+    """Return a copy of `method`, with the type of its first parameter (usually
+    self or cls) bound to original_type.
+
+    If the type of `self` is a generic type (T, or Type[T] for classmethods),
+    instantiate every occurrence of type with original_type in the rest of the
+    signature and in the return type.
+
+    original_type is the type of E in the expression E.copy(). It is None in
+    compatibility checks. In this case we treat it as the erasure of the
+    declared type of self.
+
+    This way we can express "the type of self". For example:
+
+    T = TypeVar('T', bound='A')
+    class A:
+        def copy(self: T) -> T: ...
+
+    class B(A): pass
+
+    b = B().copy()  # type: B
+
+    """
+    if isinstance(method, Overloaded):
+        return cast(F, Overloaded([bind_self(c, method) for c in method.items()]))
+    assert isinstance(method, CallableType)
+    func = method
+    if not func.arg_types:
+        # invalid method. return something
+        return cast(F, func)
+    if func.arg_kinds[0] == ARG_STAR:
+        # The signature is of the form 'def foo(*args, ...)'.
+        # In this case we shouldn't drop the first arg,
+        # since func will be absorbed by the *args.
+
+        # TODO: infer bounds on the type of *args?
+        return cast(F, func)
+    self_param_type = func.arg_types[0]
+    if func.variables and (isinstance(self_param_type, TypeVarType) or
+                           (isinstance(self_param_type, TypeType) and
+                            isinstance(self_param_type.item, TypeVarType))):
+        if original_type is None:
+            # Type check method override
+            # XXX value restriction as union?
+            original_type = erase_to_bound(self_param_type)
+
+        typearg = infer_type_arguments([x.id for x in func.variables],
+                                       self_param_type, original_type)[0]
+
+        def expand(target: Type) -> Type:
+            return expand_type(target, {func.variables[0].id: typearg})
+
+        arg_types = [expand(x) for x in func.arg_types[1:]]
+        ret_type = expand(func.ret_type)
+        variables = func.variables[1:]
+    else:
+        arg_types = func.arg_types[1:]
+        ret_type = func.ret_type
+        variables = func.variables
+    res = func.copy_modified(arg_types=arg_types,
+                             arg_kinds=func.arg_kinds[1:],
+                             arg_names=func.arg_names[1:],
+                             variables=variables,
+                             ret_type=ret_type)
+    return cast(F, res)
+
+
+def erase_to_bound(t: Type):
+    if isinstance(t, TypeVarType):
+        return t.upper_bound
+    if isinstance(t, TypeType):
+        if isinstance(t.item, TypeVarType):
+            return TypeType(t.item.upper_bound)
+    return t
