@@ -1,13 +1,13 @@
-from typing import Dict, Tuple, List, cast
+from typing import Dict, List
 
 from mypy.types import (
     Type, Instance, CallableType, TypeVisitor, UnboundType, ErrorType, AnyType,
     Void, NoneTyp, TypeVarType, Overloaded, TupleType, UnionType, ErasedType, TypeList,
-    PartialType, DeletedType, TypeType
+    PartialType, DeletedType, UninhabitedType, TypeType, TypeVarId
 )
 
 
-def expand_type(typ: Type, env: Dict[int, Type]) -> Type:
+def expand_type(typ: Type, env: Dict[TypeVarId, Type]) -> Type:
     """Substitute any type variable references in a type given by a type
     environment.
     """
@@ -16,23 +16,24 @@ def expand_type(typ: Type, env: Dict[int, Type]) -> Type:
 
 
 def expand_type_by_instance(typ: Type, instance: Instance) -> Type:
-    """Substitute type variables in type using values from an Instance."""
+    """Substitute type variables in type using values from an Instance.
+    Type variables are considered to be bound by the class declaration."""
 
     if instance.args == []:
         return typ
     else:
-        variables = {}  # type: Dict[int, Type]
-        for i in range(len(instance.args)):
-            variables[i + 1] = instance.args[i]
+        variables = {}  # type: Dict[TypeVarId, Type]
+        for binder, arg in zip(instance.type.defn.type_vars, instance.args):
+            variables[binder.id] = arg
         return expand_type(typ, variables)
 
 
 class ExpandTypeVisitor(TypeVisitor[Type]):
     """Visitor that substitutes type variables with values."""
 
-    variables = None  # type: Dict[int, Type]  # TypeVar id -> TypeVar value
+    variables = None  # type: Dict[TypeVarId, Type]  # TypeVar id -> TypeVar value
 
-    def __init__(self, variables: Dict[int, Type]) -> None:
+    def __init__(self, variables: Dict[TypeVarId, Type]) -> None:
         self.variables = variables
 
     def visit_unbound_type(self, t: UnboundType) -> Type:
@@ -53,6 +54,9 @@ class ExpandTypeVisitor(TypeVisitor[Type]):
     def visit_none_type(self, t: NoneTyp) -> Type:
         return t
 
+    def visit_uninhabited_type(self, t: UninhabitedType) -> Type:
+        return t
+
     def visit_deleted_type(self, t: DeletedType) -> Type:
         return t
 
@@ -62,14 +66,15 @@ class ExpandTypeVisitor(TypeVisitor[Type]):
 
     def visit_instance(self, t: Instance) -> Type:
         args = self.expand_types(t.args)
-        return Instance(t.type, args, t.line)
+        return Instance(t.type, args, t.line, t.column)
 
     def visit_type_var(self, t: TypeVarType) -> Type:
         repl = self.variables.get(t.id, t)
         if isinstance(repl, Instance):
-            inst = cast(Instance, repl)
+            inst = repl
             # Return copy of instance with type erasure flag on.
-            return Instance(inst.type, inst.args, inst.line, True)
+            return Instance(inst.type, inst.args, line=inst.line,
+                            column=inst.column, erased=True)
         else:
             return repl
 
@@ -80,16 +85,18 @@ class ExpandTypeVisitor(TypeVisitor[Type]):
     def visit_overloaded(self, t: Overloaded) -> Type:
         items = []  # type: List[CallableType]
         for item in t.items():
-            items.append(cast(CallableType, item.accept(self)))
+            new_item = item.accept(self)
+            assert isinstance(new_item, CallableType)
+            items.append(new_item)
         return Overloaded(items)
 
     def visit_tuple_type(self, t: TupleType) -> Type:
-        return TupleType(self.expand_types(t.items), t.fallback, t.line)
+        return t.copy_modified(items=self.expand_types(t.items))
 
     def visit_union_type(self, t: UnionType) -> Type:
         # After substituting for type variables in t.items,
         # some of the resulting types might be subtypes of others.
-        return UnionType.make_simplified_union(self.expand_types(t.items), t.line)
+        return UnionType.make_simplified_union(self.expand_types(t.items), t.line, t.column)
 
     def visit_partial_type(self, t: PartialType) -> Type:
         return t
