@@ -1,11 +1,11 @@
 """Fix up various things after deserialization."""
 
-from typing import Any, Dict, Optional, cast
+from typing import Any, Dict, Optional
 
 from mypy.nodes import (MypyFile, SymbolNode, SymbolTable, SymbolTableNode,
                         TypeInfo, FuncDef, OverloadedFuncDef, Decorator, Var,
                         TypeVarExpr, ClassDef,
-                        LDEF, MDEF, GDEF, MODULE_REF)
+                        LDEF, MDEF, GDEF)
 from mypy.types import (CallableType, EllipsisType, Instance, Overloaded, TupleType,
                         TypeList, TypeVarType, UnboundType, UnionType, TypeVisitor,
                         TypeType)
@@ -90,9 +90,6 @@ class NodeFixer(NodeVisitor[None]):
             func.info = self.current_info
         if func.type is not None:
             func.type.accept(self.type_fixer)
-        for arg in func.arguments:
-            if arg.type_annotation is not None:
-                arg.type_annotation.accept(self.type_fixer)
 
     def visit_overloaded_func_def(self, o: OverloadedFuncDef) -> None:
         if self.current_info is not None:
@@ -182,6 +179,9 @@ class TypeFixer(TypeVisitor[None]):
     def visit_none_type(self, o: Any) -> None:
         pass  # Nothing to descend into.
 
+    def visit_uninhabited_type(self, o: Any) -> None:
+        pass  # Nothing to descend into.
+
     def visit_partial_type(self, o: Any) -> None:
         raise RuntimeError("Shouldn't get here", o)
 
@@ -233,9 +233,9 @@ def lookup_qualified_stnode(modules: Dict[str, MypyFile], name: str) -> SymbolTa
     while True:
         assert '.' in head, "Cannot find %s" % (name,)
         head, tail = head.rsplit('.', 1)
+        rest.append(tail)
         mod = modules.get(head)
         if mod is not None:
-            rest.append(tail)
             break
     names = mod.names
     while True:
@@ -247,32 +247,4 @@ def lookup_qualified_stnode(modules: Dict[str, MypyFile], name: str) -> SymbolTa
             return stnode
         node = stnode.node
         assert isinstance(node, TypeInfo)
-        names = cast(TypeInfo, node).names
-
-
-def store_qualified(modules: Dict[str, MypyFile], name: str, info: SymbolNode) -> None:
-    head = name
-    rest = []
-    while True:
-        head, tail = head.rsplit('.', 1)
-        mod = modules.get(head)
-        if mod is not None:
-            rest.append(tail)
-            break
-    names = mod.names
-    while True:
-        assert rest, "Cannot find %s" % (name,)
-        key = rest.pop()
-        if key not in names:
-            assert not rest, "Cannot find %s for %s" % (key, name)
-            # Store it.
-            # TODO: kind might be something else?
-            names[key] = SymbolTableNode(GDEF, info)
-            return
-        stnode = names[key]
-        node = stnode.node
-        if not rest:
-            stnode.node = info
-            return
-        assert isinstance(node, TypeInfo)
-        names = cast(TypeInfo, node).names
+        names = node.names
