@@ -23,13 +23,9 @@ from mypy.nodes import (
 # that do not have any parameters.
 
 NO_RETURN_VALUE_EXPECTED = 'No return value expected'
+MISSING_RETURN_STATEMENT = 'Missing return statement'
 INCOMPATIBLE_RETURN_VALUE_TYPE = 'Incompatible return value type'
 RETURN_VALUE_EXPECTED = 'Return value expected'
-BOOLEAN_VALUE_EXPECTED = 'Boolean value expected'
-BOOLEAN_EXPECTED_FOR_IF = 'Boolean value expected for if condition'
-BOOLEAN_EXPECTED_FOR_WHILE = 'Boolean value expected for while condition'
-BOOLEAN_EXPECTED_FOR_UNTIL = 'Boolean value expected for until condition'
-BOOLEAN_EXPECTED_FOR_NOT = 'Boolean value expected for not operand'
 INVALID_EXCEPTION = 'Exception must be derived from BaseException'
 INVALID_EXCEPTION_TYPE = 'Exception type must be derived from BaseException'
 INVALID_RETURN_TYPE_FOR_GENERATOR = \
@@ -40,25 +36,21 @@ YIELD_VALUE_EXPECTED = 'Yield value expected'
 INCOMPATIBLE_TYPES = 'Incompatible types'
 INCOMPATIBLE_TYPES_IN_ASSIGNMENT = 'Incompatible types in assignment'
 INCOMPATIBLE_REDEFINITION = 'Incompatible redefinition'
+INCOMPATIBLE_TYPES_IN_AWAIT = 'Incompatible types in await'
+INCOMPATIBLE_TYPES_IN_ASYNC_WITH_AENTER = 'Incompatible types in "async with" for __aenter__'
+INCOMPATIBLE_TYPES_IN_ASYNC_WITH_AEXIT = 'Incompatible types in "async with" for __aexit__'
+INCOMPATIBLE_TYPES_IN_ASYNC_FOR = 'Incompatible types in "async for"'
+
 INCOMPATIBLE_TYPES_IN_YIELD = 'Incompatible types in yield'
 INCOMPATIBLE_TYPES_IN_YIELD_FROM = 'Incompatible types in "yield from"'
 INCOMPATIBLE_TYPES_IN_STR_INTERPOLATION = 'Incompatible types in string interpolation'
 INIT_MUST_HAVE_NONE_RETURN_TYPE = 'The return type of "__init__" must be None'
-GETTER_TYPE_INCOMPATIBLE_WITH_SETTER = \
-    'Type of getter incompatible with setter'
 TUPLE_INDEX_MUST_BE_AN_INT_LITERAL = 'Tuple index must be an integer literal'
 TUPLE_SLICE_MUST_BE_AN_INT_LITERAL = 'Tuple slice must be an integer literal'
 TUPLE_INDEX_OUT_OF_RANGE = 'Tuple index out of range'
-TYPE_CONSTANT_EXPECTED = 'Type "Constant" or initializer expected'
-INCOMPATIBLE_PAIR_ITEM_TYPE = 'Incompatible Pair item type'
-INVALID_TYPE_APPLICATION_TARGET_TYPE = 'Invalid type application target type'
-INCOMPATIBLE_TUPLE_ITEM_TYPE = 'Incompatible tuple item type'
-INCOMPATIBLE_KEY_TYPE = 'Incompatible dictionary key type'
-INCOMPATIBLE_VALUE_TYPE = 'Incompatible dictionary value type'
 NEED_ANNOTATION_FOR_VAR = 'Need type annotation for variable'
 ITERABLE_EXPECTED = 'Iterable expected'
-INCOMPATIBLE_TYPES_IN_FOR = 'Incompatible types in for statement'
-INCOMPATIBLE_ARRAY_VAR_ARGS = 'Incompatible variable arguments in call'
+ASYNC_ITERABLE_EXPECTED = 'AsyncIterable expected'
 INVALID_SLICE_INDEX = 'Slice index must be an integer or None'
 CANNOT_INFER_LAMBDA_TYPE = 'Cannot infer type of lambda'
 CANNOT_INFER_ITEM_TYPE = 'Cannot infer iterable item type'
@@ -69,14 +61,12 @@ INCONSISTENT_ABSTRACT_OVERLOAD = \
     'Overloaded method has both abstract and non-abstract variants'
 READ_ONLY_PROPERTY_OVERRIDES_READ_WRITE = \
     'Read-only property cannot override read-write property'
-INSTANCE_LAYOUT_CONFLICT = 'Instance layout conflict in multiple inheritance'
 FORMAT_REQUIRES_MAPPING = 'Format requires a mapping'
-GENERIC_TYPE_NOT_VALID_AS_EXPRESSION = \
-    "Generic type is prohibited as a runtime expression (use a type alias or '# type:' comment)"
 RETURN_TYPE_CANNOT_BE_CONTRAVARIANT = "Cannot use a contravariant type variable as return type"
 FUNCTION_PARAMETER_CANNOT_BE_COVARIANT = "Cannot use a covariant type variable as a parameter"
 INCOMPATIBLE_IMPORT_OF = "Incompatible import of"
 FUNCTION_TYPE_EXPECTED = "Function is missing a type annotation"
+ONLY_CLASS_APPLICATION = "Type application is only supported for generic classes"
 RETURN_TYPE_EXPECTED = "Function is missing a return type annotation"
 ARGUMENT_TYPE_EXPECTED = "Function is missing a type annotation for one or more arguments"
 KEYWORD_ARGUMENT_REQUIRES_STR_KEY_TYPE = \
@@ -142,6 +132,7 @@ class MessageBuilder:
         """Report an error or note (unless disabled)."""
         if self.disable_count <= 0:
             self.errors.report(context.get_line() if context else -1,
+                               context.get_column() if context else -1,
                                msg.strip(), severity=severity, file=file)
 
     def fail(self, msg: str, context: Context, file: str = None) -> None:
@@ -229,7 +220,7 @@ class MessageBuilder:
                 for arg in itype.args:
                     a.append(strip_quotes(self.format(arg)))
                 s = ', '.join(a)
-                if len((base_str + s)) < 25:
+                if len((base_str + s)) < 150:
                     return '{}[{}]'.format(base_str, s)
                 else:
                     return '{}[...]'.format(base_str)
@@ -244,7 +235,7 @@ class MessageBuilder:
             for t in typ.items:
                 items.append(strip_quotes(self.format(t)))
             s = '"Tuple[{}]"'.format(', '.join(items))
-            if len(s) < 40:
+            if len(s) < 400:
                 return s
             else:
                 return 'tuple(length {})'.format(len(items))
@@ -260,7 +251,7 @@ class MessageBuilder:
                 for t in typ.items:
                     items.append(strip_quotes(self.format(t)))
                 s = '"Union[{}]"'.format(', '.join(items))
-                if len(s) < 40:
+                if len(s) < 400:
                     return s
                 else:
                     return 'union type ({} items)'.format(len(items))
@@ -314,8 +305,8 @@ class MessageBuilder:
         if (isinstance(typ, Instance) and
                 typ.type.has_readable_member(member)):
             self.fail('Member "{}" is not assignable'.format(member), context)
-        elif isinstance(typ, Void):
-            self.check_void(typ, context)
+        elif self.check_unusable_type(typ, context):
+            pass
         elif member == '__contains__':
             self.fail('Unsupported right operand type for in ({})'.format(
                 self.format(typ)), context)
@@ -343,7 +334,12 @@ class MessageBuilder:
             # Indexed set.
             self.fail('Unsupported target for indexed assignment', context)
         elif member == '__call__':
-            self.fail('{} not callable'.format(self.format(typ)), context)
+            if isinstance(typ, Instance) and (typ.type.fullname() == 'builtins.function'):
+                # "'function' not callable" is a confusing error message.
+                # Explain that the problem is that the type of the function is not known.
+                self.fail('Cannot call function of unknown type', context)
+            else:
+                self.fail('{} not callable'.format(self.format(typ)), context)
         else:
             # The non-special case: a missing ordinary attribute.
             if not self.disable_type_names:
@@ -370,9 +366,8 @@ class MessageBuilder:
 
         Types can be Type objects or strings.
         """
-        if isinstance(left_type, Void) or isinstance(right_type, Void):
-            self.check_void(left_type, context)
-            self.check_void(right_type, context)
+        if (self.check_unusable_type(left_type, context) or
+                self.check_unusable_type(right_type, context)):
             return
         left_str = ''
         if isinstance(left_type, str):
@@ -395,7 +390,7 @@ class MessageBuilder:
 
     def unsupported_left_operand(self, op: str, typ: Type,
                                  context: Context) -> None:
-        if not self.check_void(typ, context):
+        if not self.check_unusable_type(typ, context):
             if self.disable_type_names:
                 msg = 'Unsupported left operand type for {} (some union)'.format(op)
             else:
@@ -403,16 +398,13 @@ class MessageBuilder:
                     op, self.format(typ))
             self.fail(msg, context)
 
-    def type_expected_as_right_operand_of_is(self, context: Context) -> None:
-        self.fail('Type expected as right operand of "is"', context)
-
     def not_callable(self, typ: Type, context: Context) -> Type:
         self.fail('{} not callable'.format(self.format(typ)), context)
         return AnyType()
 
     def untyped_function_call(self, callee: CallableType, context: Context) -> Type:
         name = callee.name if callee.name is not None else '(unknown)'
-        self.fail('call to untyped function {} in typed context'.format(name), context)
+        self.fail('Call to untyped function {} in typed context'.format(name), context)
         return AnyType()
 
     def incompatible_argument(self, n: int, m: int, callee: CallableType, arg_type: Type,
@@ -548,18 +540,17 @@ class MessageBuilder:
                   format(capitalize(callable_name(callee)),
                          callee.arg_names[index]), context)
 
-    def does_not_return_value(self, void_type: Type, context: Context) -> None:
-        """Report an error about a void type in a non-void context.
+    def does_not_return_value(self, unusable_type: Type, context: Context) -> None:
+        """Report an error about use of an unusable type.
 
-        The first argument must be a void type. If the void type has a
-        source in it, report it in the error message. This allows
-        giving messages such as 'Foo does not return a value'.
+        If the type is a Void type and has a source in it, report it in the error message.
+        This allows giving messages such as 'Foo does not return a value'.
         """
-        if (cast(Void, void_type)).source is None:
-            self.fail('Function does not return a value', context)
-        else:
+        if isinstance(unusable_type, Void) and unusable_type.source is not None:
             self.fail('{} does not return a value'.format(
-                capitalize((cast(Void, void_type)).source)), context)
+                capitalize((cast(Void, unusable_type)).source)), context)
+        else:
+            self.fail('Function does not return a value', context)
 
     def deleted_as_rvalue(self, typ: DeletedType, context: Context) -> None:
         """Report an error about using an deleted type as an rvalue."""
@@ -589,14 +580,9 @@ class MessageBuilder:
         else:
             self.fail('No overload variant matches argument types {}'.format(arg_types), context)
 
-    def function_variants_overlap(self, n1: int, n2: int,
-                                  context: Context) -> None:
-        self.fail('Function signature variants {} and {} overlap'.format(
-            n1 + 1, n2 + 1), context)
-
     def invalid_cast(self, target_type: Type, source_type: Type,
                      context: Context) -> None:
-        if not self.check_void(source_type, context):
+        if not self.check_unusable_type(source_type, context):
             self.fail('Cannot cast from {} to {}'.format(
                 self.format(source_type), self.format(target_type)), context)
 
@@ -621,19 +607,6 @@ class MessageBuilder:
         self.fail('Result type of {} incompatible in assignment'.format(op),
                   context)
 
-    def incompatible_value_count_in_assignment(self, lvalue_count: int,
-                                               rvalue_count: int,
-                                               context: Context) -> None:
-        if rvalue_count < lvalue_count:
-            self.fail('Need {} values to assign'.format(lvalue_count), context)
-        elif rvalue_count > lvalue_count:
-            self.fail('Too many values to assign', context)
-
-    def type_incompatible_with_supertype(self, name: str, supertype: TypeInfo,
-                                         context: Context) -> None:
-        self.fail('Type of "{}" incompatible with supertype "{}"'.format(
-            name, supertype.name), context)
-
     def signature_incompatible_with_supertype(
             self, name: str, name_in_super: str, supertype: str,
             context: Context) -> None:
@@ -662,16 +635,11 @@ class MessageBuilder:
             target = '"{}" of {}'.format(name_in_super, target)
         return target
 
-    def boolean_return_value_expected(self, method: str,
-                                      context: Context) -> None:
-        self.fail('Boolean return value expected for method "{}"'.format(
-            method), context)
-
     def incompatible_type_application(self, expected_arg_count: int,
                                       actual_arg_count: int,
                                       context: Context) -> None:
         if expected_arg_count == 0:
-            self.fail('Type application targets a non-generic function',
+            self.fail('Type application targets a non-generic function or class',
                       context)
         elif actual_arg_count > expected_arg_count:
             self.fail('Type application has too many types ({} expected)'
@@ -680,11 +648,6 @@ class MessageBuilder:
             self.fail('Type application has too few types ({} expected)'
                       .format(expected_arg_count), context)
 
-    def incompatible_array_item_type(self, typ: Type, index: int,
-                                     context: Context) -> None:
-        self.fail('Array item {} has incompatible type {}'.format(
-            index, self.format(typ)), context)
-
     def could_not_infer_type_arguments(self, callee_type: CallableType, n: int,
                                        context: Context) -> None:
         if callee_type.name and n > 0:
@@ -703,22 +666,16 @@ class MessageBuilder:
             self.fail('Argument after ** must be a dictionary',
                       context)
 
-    def incomplete_type_var_match(self, member: str, context: Context) -> None:
-        self.fail('"{}" has incomplete match to supertype type variable'
-                  .format(member), context)
-
-    def not_implemented(self, msg: str, context: Context) -> Type:
-        self.fail('Feature not implemented yet ({})'.format(msg), context)
-        return AnyType()
-
     def undefined_in_superclass(self, member: str, context: Context) -> None:
         self.fail('"{}" undefined in superclass'.format(member), context)
 
-    def check_void(self, typ: Type, context: Context) -> bool:
-        """If type is void, report an error such as '.. does not
+    def check_unusable_type(self, typ: Type, context: Context) -> bool:
+        """If type is a type which is not meant to be used (like Void or
+        NoneTyp(is_ret_type=True)), report an error such as '.. does not
         return a value' and return True. Otherwise, return False.
         """
-        if isinstance(typ, Void):
+        if (isinstance(typ, Void) or
+                (isinstance(typ, NoneTyp) and typ.is_ret_type)):
             self.does_not_return_value(typ, context)
             return True
         else:
@@ -730,9 +687,6 @@ class MessageBuilder:
     def too_many_string_formatting_arguments(self, context: Context) -> None:
         self.fail('Not all arguments converted during string formatting', context)
 
-    def incomplete_conversion_specifier_format(self, context: Context) -> None:
-        self.fail('Incomplete format', context)
-
     def unsupported_placeholder(self, placeholder: str, context: Context) -> None:
         self.fail('Unsupported format character \'%s\'' % placeholder, context)
 
@@ -767,7 +721,7 @@ class MessageBuilder:
     def cannot_instantiate_abstract_class(self, class_name: str,
                                           abstract_attributes: List[str],
                                           context: Context) -> None:
-        attrs = format_string_list("'%s'" % a for a in abstract_attributes[:5])
+        attrs = format_string_list("'%s'" % a for a in abstract_attributes)
         self.fail("Cannot instantiate abstract class '%s' with abstract "
                   "attribute%s %s" % (class_name, plural_s(abstract_attributes),
                                    attrs),
@@ -798,16 +752,6 @@ class MessageBuilder:
         self.fail('Overloaded function signatures {} and {} overlap with '
                   'incompatible return types'.format(index1, index2), context)
 
-    def invalid_reverse_operator_signature(self, reverse: str, other: str,
-                                           context: Context) -> None:
-        self.fail('"Any" return type expected since argument to {} '
-                  'does not support {}'.format(reverse, other), context)
-
-    def reverse_operator_method_with_any_arg_must_return_any(
-            self, method: str, context: Context) -> None:
-        self.fail('"Any" return type expected since argument to {} '
-                  'has type "Any"'.format(method), context)
-
     def operator_method_signatures_overlap(
             self, reverse_class: str, reverse_method: str, forward_class: str,
             forward_method: str, context: Context) -> None:
@@ -881,8 +825,10 @@ def format_string_list(s: Iterable[str]) -> str:
     assert len(l) > 0
     if len(l) == 1:
         return l[0]
-    else:
+    elif len(l) <= 5:
         return '%s and %s' % (', '.join(l[:-1]), l[-1])
+    else:
+        return '%s, ... and %s (%i methods suppressed)' % (', '.join(l[:2]), l[-1], len(l) - 3)
 
 
 def callable_name(type: CallableType) -> str:
