@@ -1,15 +1,15 @@
 """Semantic analysis of types"""
 
-from typing import Callable, cast, List, Tuple
+from typing import Callable, cast, List, Optional
 
 from mypy.types import (
-    Type, UnboundType, TypeVarType, TupleType, UnionType, Instance, AnyType, CallableType,
-    Void, NoneTyp, DeletedType, TypeList, TypeVarDef, TypeVisitor, StarType, PartialType,
-    EllipsisType, UninhabitedType, TypeType
+    Type, UnboundType, TypeVarType, TupleType, UnionType, Instance,
+    AnyType, CallableType, Void, NoneTyp, DeletedType, TypeList, TypeVarDef, TypeVisitor,
+    StarType, PartialType, EllipsisType, UninhabitedType, TypeType, get_typ_args, set_typ_args,
 )
 from mypy.nodes import (
-    BOUND_TVAR, TYPE_ALIAS, UNBOUND_IMPORTED,
-    TypeInfo, Context, SymbolTableNode, TypeVarExpr, Var, Node,
+    BOUND_TVAR, UNBOUND_TVAR, TYPE_ALIAS, UNBOUND_IMPORTED,
+    TypeInfo, Context, SymbolTableNode, Var, Expression,
     IndexExpr, RefExpr
 )
 from mypy.sametypes import is_same_type
@@ -19,10 +19,16 @@ from mypy import nodes
 from mypy import experiments
 
 
-type_constructors = {'typing.Tuple', 'typing.Union', 'typing.Callable', 'typing.Type'}
+type_constructors = {
+    'typing.Callable',
+    'typing.Optional',
+    'typing.Tuple',
+    'typing.Type',
+    'typing.Union',
+}
 
 
-def analyze_type_alias(node: Node,
+def analyze_type_alias(node: Expression,
                        lookup_func: Callable[[str, Context], SymbolTableNode],
                        lookup_fqn_func: Callable[[str], SymbolTableNode],
                        fail_func: Callable[[str, Context], None]) -> Type:
@@ -34,6 +40,10 @@ def analyze_type_alias(node: Node,
     # that we don't support straight string literals as type aliases
     # (only string literals within index expressions).
     if isinstance(node, RefExpr):
+        if node.kind == UNBOUND_TVAR or node.kind == BOUND_TVAR:
+            fail_func('Type variable "{}" is invalid as target for type alias'.format(
+                node.fullname), node)
+            return None
         if not (isinstance(node.node, TypeInfo) or
                 node.fullname == 'typing.Any' or
                 node.kind == TYPE_ALIAS):
@@ -42,7 +52,8 @@ def analyze_type_alias(node: Node,
         base = node.base
         if isinstance(base, RefExpr):
             if not (isinstance(base.node, TypeInfo) or
-                    base.fullname in type_constructors):
+                    base.fullname in type_constructors or
+                    base.kind == TYPE_ALIAS):
                 return None
         else:
             return None
@@ -55,7 +66,7 @@ def analyze_type_alias(node: Node,
     except TypeTranslationError:
         fail_func('Invalid type alias', node)
         return None
-    analyzer = TypeAnalyser(lookup_func, lookup_fqn_func, fail_func)
+    analyzer = TypeAnalyser(lookup_func, lookup_fqn_func, fail_func, aliasing=True)
     return type.accept(analyzer)
 
 
@@ -68,10 +79,12 @@ class TypeAnalyser(TypeVisitor[Type]):
     def __init__(self,
                  lookup_func: Callable[[str, Context], SymbolTableNode],
                  lookup_fqn_func: Callable[[str], SymbolTableNode],
-                 fail_func: Callable[[str, Context], None]) -> None:
+                 fail_func: Callable[[str, Context], None], *,
+                 aliasing: bool = False) -> None:
         self.lookup = lookup_func
         self.lookup_fqn_func = lookup_fqn_func
         self.fail = fail_func
+        self.aliasing = aliasing
 
     def visit_unbound_type(self, t: UnboundType) -> Type:
         if t.optional:
@@ -95,20 +108,20 @@ class TypeAnalyser(TypeVisitor[Type]):
                 return TypeVarType(sym.tvar_def, t.line)
             elif fullname == 'builtins.None':
                 if experiments.STRICT_OPTIONAL:
-                    if t.is_ret_type:
-                        return Void()
-                    else:
-                        return NoneTyp()
+                    return NoneTyp(is_ret_type=t.is_ret_type)
                 else:
                     return Void()
             elif fullname == 'typing.Any':
                 return AnyType()
             elif fullname == 'typing.Tuple':
+                if len(t.args) == 0 and not t.empty_tuple_index:
+                    # Bare 'Tuple' is same as 'tuple'
+                    return self.builtin_type('builtins.tuple')
                 if len(t.args) == 2 and isinstance(t.args[1], EllipsisType):
                     # Tuple[T, ...] (uniform, variable-length tuple)
-                    node = self.lookup_fqn_func('builtins.tuple')
-                    tuple_info = cast(TypeInfo, node.node)
-                    return Instance(tuple_info, [t.args[0].accept(self)], t.line)
+                    instance = self.builtin_type('builtins.tuple', [t.args[0].accept(self)])
+                    instance.line = t.line
+                    return instance
                 return self.tuple_type(self.anal_array(t.args))
             elif fullname == 'typing.Union':
                 items = self.anal_array(t.args)
@@ -135,8 +148,22 @@ class TypeAnalyser(TypeVisitor[Type]):
                 item = items[0]
                 return TypeType(item, line=t.line)
             elif sym.kind == TYPE_ALIAS:
-                # TODO: Generic type aliases.
-                return sym.type_override
+                override = sym.type_override
+                an_args = self.anal_array(t.args)
+                all_vars = self.get_type_var_names(override)
+                exp_len = len(all_vars)
+                act_len = len(an_args)
+                if exp_len > 0 and act_len == 0:
+                    # Interpret bare Alias same as normal generic, i.e., Alias[Any, Any, ...]
+                    return self.replace_alias_tvars(override, all_vars, [AnyType()] * exp_len,
+                                                    t.line, t.column)
+                if exp_len == 0 and act_len == 0:
+                    return override
+                if act_len != exp_len:
+                    self.fail('Bad number of arguments for type alias, expected: %s, given: %s'
+                              % (exp_len, act_len), t)
+                    return t
+                return self.replace_alias_tvars(override, all_vars, an_args, t.line, t.column)
             elif not isinstance(sym.node, TypeInfo):
                 name = sym.fullname
                 if name is None:
@@ -147,7 +174,9 @@ class TypeAnalyser(TypeVisitor[Type]):
                     # as a base class -- however, this will fail soon at runtime so the problem
                     # is pretty minor.
                     return AnyType()
-                self.fail('Invalid type "{}"'.format(name), t)
+                # Allow unbound type variables when defining an alias
+                if not (self.aliasing and sym.kind == UNBOUND_TVAR):
+                    self.fail('Invalid type "{}"'.format(name), t)
                 return t
             info = sym.node  # type: TypeInfo
             if len(t.args) > 0 and info.fullname() == 'builtins.tuple':
@@ -160,8 +189,9 @@ class TypeAnalyser(TypeVisitor[Type]):
                 # checked only later, since we do not always know the
                 # valid count at this point. Thus we may construct an
                 # Instance with an invalid number of type arguments.
-                instance = Instance(info, self.anal_array(t.args), t.line)
-                if info.tuple_type is None:
+                instance = Instance(info, self.anal_array(t.args), t.line, t.column)
+                tup = info.tuple_type
+                if tup is None:
                     return instance
                 else:
                     # The class has a Tuple[...] base class so it will be
@@ -169,12 +199,59 @@ class TypeAnalyser(TypeVisitor[Type]):
                     if t.args:
                         self.fail('Generic tuple types not supported', t)
                         return AnyType()
-                    return TupleType(self.anal_array(info.tuple_type.items),
-                                     fallback=instance,
-                                     line=t.line)
+                    return tup.copy_modified(items=self.anal_array(tup.items),
+                                             fallback=instance)
         else:
             return AnyType()
 
+    def get_type_var_names(self, tp: Type) -> List[str]:
+        """Get all type variable names that are present in a generic type alias
+        in order of textual appearance (recursively, if needed).
+        """
+        tvars = []  # type: List[str]
+        typ_args = get_typ_args(tp)
+        for arg in typ_args:
+            tvar = self.get_tvar_name(arg)
+            if tvar:
+                tvars.append(tvar)
+            else:
+                subvars = self.get_type_var_names(arg)
+                if subvars:
+                    tvars.extend(subvars)
+        # Get unique type variables in order of appearance
+        all_tvars = set(tvars)
+        new_tvars = []
+        for t in tvars:
+            if t in all_tvars:
+                new_tvars.append(t)
+                all_tvars.remove(t)
+        return new_tvars
+
+    def get_tvar_name(self, t: Type) -> Optional[str]:
+        if not isinstance(t, UnboundType):
+            return None
+        sym = self.lookup(t.name, t)
+        if sym is not None and (sym.kind == UNBOUND_TVAR or sym.kind == BOUND_TVAR):
+            return t.name
+        return None
+
+    def replace_alias_tvars(self, tp: Type, vars: List[str], subs: List[Type],
+                            newline: int, newcolumn: int) -> Type:
+        """Replace type variables in a generic type alias tp with substitutions subs
+        resetting context. Length of subs should be already checked.
+        """
+        typ_args = get_typ_args(tp)
+        new_args = typ_args[:]
+        for i, arg in enumerate(typ_args):
+            tvar = self.get_tvar_name(arg)
+            if tvar and tvar in vars:
+                # Perform actual substitution...
+                new_args[i] = subs[vars.index(tvar)]
+            else:
+                # ...recursively, if needed.
+                new_args[i] = self.replace_alias_tvars(arg, vars, subs, newline, newcolumn)
+        return set_typ_args(tp, new_args, newline, newcolumn)
+
     def visit_any(self, t: AnyType) -> Type:
         return t
 
@@ -198,7 +275,7 @@ class TypeAnalyser(TypeVisitor[Type]):
         return t
 
     def visit_type_var(self, t: TypeVarType) -> Type:
-        raise RuntimeError('TypeVarType is already analyzed')
+        return t
 
     def visit_callable_type(self, t: CallableType) -> Type:
         return t.copy_modified(arg_types=self.anal_array(t.arg_types),
@@ -285,8 +362,8 @@ class TypeAnalyser(TypeVisitor[Type]):
 
     def builtin_type(self, fully_qualified_name: str, args: List[Type] = None) -> Instance:
         node = self.lookup_fqn_func(fully_qualified_name)
-        info = cast(TypeInfo, node.node)
-        return Instance(info, args or [])
+        assert isinstance(node.node, TypeInfo)
+        return Instance(node.node, args or [])
 
     def tuple_type(self, items: List[Type]) -> TupleType:
         return TupleType(items, fallback=self.builtin_type('builtins.tuple', [AnyType()]))
@@ -339,9 +416,10 @@ class TypeAnalyserPass3(TypeVisitor[None]):
             # otherwise the type checker may crash as it expects
             # things to be right.
             t.args = [AnyType() for _ in info.type_vars]
+            t.invalid = True
         elif info.defn.type_vars:
             # Check type argument values.
-            for arg, TypeVar in zip(t.args, info.defn.type_vars):
+            for (i, arg), TypeVar in zip(enumerate(t.args), info.defn.type_vars):
                 if TypeVar.values:
                     if isinstance(arg, TypeVarType):
                         arg_values = arg.values
@@ -353,7 +431,7 @@ class TypeAnalyserPass3(TypeVisitor[None]):
                     else:
                         arg_values = [arg]
                     self.check_type_var_values(info, arg_values,
-                                               TypeVar.values, t)
+                                               TypeVar.values, i + 1, t)
                 if not satisfies_upper_bound(arg, TypeVar.upper_bound):
                     self.fail('Type argument "{}" of "{}" must be '
                               'a subtype of "{}"'.format(
@@ -362,12 +440,16 @@ class TypeAnalyserPass3(TypeVisitor[None]):
             arg.accept(self)
 
     def check_type_var_values(self, type: TypeInfo, actuals: List[Type],
-                              valids: List[Type], context: Context) -> None:
+                              valids: List[Type], arg_number: int, context: Context) -> None:
         for actual in actuals:
             if (not isinstance(actual, AnyType) and
                     not any(is_same_type(actual, value) for value in valids)):
-                self.fail('Invalid type argument value for "{}"'.format(
-                    type.name()), context)
+                if len(actuals) > 1 or not isinstance(actual, Instance):
+                    self.fail('Invalid type argument value for "{}"'.format(
+                        type.name()), context)
+                else:
+                    self.fail('Type argument {} of "{}" has incompatible value "{}"'.format(
+                        arg_number, type.name(), actual.type.name()), context)
 
     def visit_callable_type(self, t: CallableType) -> None:
         t.ret_type.accept(self)
