@@ -1,6 +1,7 @@
 """Classes for representing mypy types."""
 
 from abc import abstractmethod
+import copy
 from typing import (
     Any, TypeVar, Dict, List, Tuple, cast, Generic, Set, Sequence, Optional, Union
 )
@@ -20,13 +21,20 @@ class Type(mypy.nodes.Context):
     """Abstract base class for all types."""
 
     line = 0
+    column = 0
+    can_be_true = True
+    can_be_false = True
 
-    def __init__(self, line: int = -1) -> None:
+    def __init__(self, line: int = -1, column: int = -1) -> None:
         self.line = line
+        self.column = column
 
     def get_line(self) -> int:
         return self.line
 
+    def get_column(self) -> int:
+        return self.column
+
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         raise RuntimeError('Not implemented')
 
@@ -107,9 +115,11 @@ class TypeVarDef(mypy.nodes.Context):
     upper_bound = None  # type: Type
     variance = INVARIANT  # type: int
     line = 0
+    column = 0
 
     def __init__(self, name: str, id: Union[TypeVarId, int], values: Optional[List[Type]],
-                 upper_bound: Type, variance: int = INVARIANT, line: int = -1) -> None:
+                 upper_bound: Type, variance: int = INVARIANT, line: int = -1,
+                 column: int = -1) -> None:
         self.name = name
         if isinstance(id, int):
             id = TypeVarId(id)
@@ -118,16 +128,20 @@ class TypeVarDef(mypy.nodes.Context):
         self.upper_bound = upper_bound
         self.variance = variance
         self.line = line
+        self.column = column
 
     @staticmethod
     def new_unification_variable(old: 'TypeVarDef') -> 'TypeVarDef':
         new_id = TypeVarId.new(meta_level=1)
         return TypeVarDef(old.name, new_id, old.values,
-                          old.upper_bound, old.variance, old.line)
+                          old.upper_bound, old.variance, old.line, old.column)
 
     def get_line(self) -> int:
         return self.line
 
+    def get_column(self) -> int:
+        return self.column
+
     def __repr__(self) -> str:
         if self.values:
             return '{} in {}'.format(self.name, tuple(self.values))
@@ -168,19 +182,25 @@ class UnboundType(Type):
     # is this type a return type?
     is_ret_type = False
 
+    # special case for X[()]
+    empty_tuple_index = False
+
     def __init__(self,
                  name: str,
                  args: List[Type] = None,
                  line: int = -1,
+                 column: int = -1,
                  optional: bool = False,
-                 is_ret_type: bool = False) -> None:
+                 is_ret_type: bool = False,
+                 empty_tuple_index: bool = False) -> None:
         if not args:
             args = []
         self.name = name
         self.args = args
         self.optional = optional
         self.is_ret_type = is_ret_type
-        super().__init__(line)
+        self.empty_tuple_index = empty_tuple_index
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_unbound_type(self)
@@ -192,7 +212,7 @@ class UnboundType(Type):
                 }
 
     @classmethod
-    def deserialize(self, data: JsonDict) -> 'UnboundType':
+    def deserialize(cls, data: JsonDict) -> 'UnboundType':
         assert data['.class'] == 'UnboundType'
         return UnboundType(data['name'],
                            [Type.deserialize(a) for a in data['args']])
@@ -215,8 +235,8 @@ class TypeList(Type):
 
     items = None  # type: List[Type]
 
-    def __init__(self, items: List[Type], line: int = -1) -> None:
-        super().__init__(line)
+    def __init__(self, items: List[Type], line: int = -1, column: int = -1) -> None:
+        super().__init__(line, column)
         self.items = items
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
@@ -228,7 +248,7 @@ class TypeList(Type):
                 }
 
     @classmethod
-    def deserialize(self, data: JsonDict) -> 'TypeList':
+    def deserialize(cls, data: JsonDict) -> 'TypeList':
         assert data['.class'] == 'TypeList'
         return TypeList([Type.deserialize(t) for t in data['items']])
 
@@ -236,8 +256,8 @@ class TypeList(Type):
 class AnyType(Type):
     """The type 'Any'."""
 
-    def __init__(self, implicit: bool = False, line: int = -1) -> None:
-        super().__init__(line)
+    def __init__(self, implicit: bool = False, line: int = -1, column: int = -1) -> None:
+        super().__init__(line, column)
         self.implicit = implicit
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
@@ -259,17 +279,18 @@ class Void(Type):
     the result type of calling such callable.
     """
 
+    can_be_true = False
     source = ''   # May be None; function that generated this value
 
-    def __init__(self, source: str = None, line: int = -1) -> None:
+    def __init__(self, source: str = None, line: int = -1, column: int = -1) -> None:
         self.source = source
-        super().__init__(line)
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_void(self)
 
     def with_source(self, source: str) -> 'Void':
-        return Void(source, self.line)
+        return Void(source, self.line, self.column)
 
     def serialize(self) -> JsonDict:
         return {'.class': 'Void'}
@@ -294,8 +315,11 @@ class UninhabitedType(Type):
         is_subtype(UninhabitedType, T) = True
     """
 
-    def __init__(self, line: int = -1) -> None:
-        super().__init__(line)
+    can_be_true = False
+    can_be_false = False
+
+    def __init__(self, line: int = -1, column: int = -1) -> None:
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_uninhabited_type(self)
@@ -327,19 +351,24 @@ class NoneTyp(Type):
         of a function, where 'None' means Void.
     """
 
-    def __init__(self, line: int = -1) -> None:
-        super().__init__(line)
+    can_be_true = False
+
+    def __init__(self, is_ret_type: bool = False, line: int = -1, column: int = -1) -> None:
+        super().__init__(line, column)
+        self.is_ret_type = is_ret_type
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_none_type(self)
 
     def serialize(self) -> JsonDict:
-        return {'.class': 'NoneTyp'}
+        return {'.class': 'NoneTyp',
+                'is_ret_type': self.is_ret_type,
+                }
 
     @classmethod
-    def deserialize(self, data: JsonDict) -> 'NoneTyp':
+    def deserialize(cls, data: JsonDict) -> 'NoneTyp':
         assert data['.class'] == 'NoneTyp'
-        return NoneTyp()
+        return NoneTyp(is_ret_type=data['is_ret_type'])
 
 
 class ErasedType(Type):
@@ -361,9 +390,9 @@ class DeletedType(Type):
 
     source = ''   # May be None; name that generated this value
 
-    def __init__(self, source: str = None, line: int = -1) -> None:
+    def __init__(self, source: str = None, line: int = -1, column: int = -1) -> None:
         self.source = source
-        super().__init__(line)
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_deleted_type(self)
@@ -373,7 +402,7 @@ class DeletedType(Type):
                 'source': self.source}
 
     @classmethod
-    def deserialize(self, data: JsonDict) -> 'DeletedType':
+    def deserialize(cls, data: JsonDict) -> 'DeletedType':
         assert data['.class'] == 'DeletedType'
         return DeletedType(data['source'])
 
@@ -386,14 +415,15 @@ class Instance(Type):
 
     type = None  # type: mypy.nodes.TypeInfo
     args = None  # type: List[Type]
-    erased = False      # True if result of type variable substitution
+    erased = False  # True if result of type variable substitution
+    invalid = False  # True if recovered after incorrect number of type arguments error
 
     def __init__(self, typ: mypy.nodes.TypeInfo, args: List[Type],
-                 line: int = -1, erased: bool = False) -> None:
+                 line: int = -1, column: int = -1, erased: bool = False) -> None:
         self.type = typ
         self.args = args
         self.erased = erased
-        super().__init__(line)
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_instance(self)
@@ -436,13 +466,13 @@ class TypeVarType(Type):
     # See comments in TypeVarDef for more about variance.
     variance = INVARIANT  # type: int
 
-    def __init__(self, binder: TypeVarDef, line: int = -1) -> None:
+    def __init__(self, binder: TypeVarDef, line: int = -1, column: int = -1) -> None:
         self.name = binder.name
         self.id = binder.id
         self.values = binder.values
         self.upper_bound = binder.upper_bound
         self.variance = binder.variance
-        super().__init__(line)
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_type_var(self)
@@ -477,6 +507,8 @@ class TypeVarType(Type):
 class FunctionLike(Type):
     """Abstract base class for function types."""
 
+    can_be_false = False
+
     @abstractmethod
     def is_type_obj(self) -> bool: pass
 
@@ -530,13 +562,14 @@ class CallableType(FunctionLike):
     def __init__(self,
                  arg_types: List[Type],
                  arg_kinds: List[int],
-                 arg_names: List[str],
+                 arg_names: List[Optional[str]],
                  ret_type: Type,
                  fallback: Instance,
                  name: str = None,
                  definition: SymbolNode = None,
                  variables: List[TypeVarDef] = None,
                  line: int = -1,
+                 column: int = -1,
                  is_ellipsis_args: bool = False,
                  implicit: bool = False,
                  is_classmethod_class: bool = False,
@@ -544,6 +577,7 @@ class CallableType(FunctionLike):
                  ) -> None:
         if variables is None:
             variables = []
+        assert len(arg_types) == len(arg_kinds)
         self.arg_types = arg_types
         self.arg_kinds = arg_kinds
         self.arg_names = arg_names
@@ -558,7 +592,7 @@ class CallableType(FunctionLike):
         self.is_ellipsis_args = is_ellipsis_args
         self.implicit = implicit
         self.special_sig = special_sig
-        super().__init__(line)
+        super().__init__(line, column)
 
     def copy_modified(self,
                       arg_types: List[Type] = _dummy,
@@ -570,6 +604,7 @@ class CallableType(FunctionLike):
                       definition: SymbolNode = _dummy,
                       variables: List[TypeVarDef] = _dummy,
                       line: int = _dummy,
+                      column: int = _dummy,
                       is_ellipsis_args: bool = _dummy,
                       special_sig: Optional[str] = _dummy) -> 'CallableType':
         return CallableType(
@@ -582,6 +617,7 @@ class CallableType(FunctionLike):
             definition=definition if definition is not _dummy else self.definition,
             variables=variables if variables is not _dummy else self.variables,
             line=line if line is not _dummy else self.line,
+            column=column if column is not _dummy else self.column,
             is_ellipsis_args=(
                 is_ellipsis_args if is_ellipsis_args is not _dummy else self.is_ellipsis_args),
             implicit=self.implicit,
@@ -680,7 +716,7 @@ class Overloaded(FunctionLike):
     def __init__(self, items: List[CallableType]) -> None:
         self._items = items
         self.fallback = items[0].fallback
-        super().__init__(items[0].line)
+        super().__init__(items[0].line, items[0].column)
 
     def items(self) -> List[CallableType]:
         return self._items
@@ -713,7 +749,7 @@ class Overloaded(FunctionLike):
                 }
 
     @classmethod
-    def deserialize(self, data: JsonDict) -> 'Overloaded':
+    def deserialize(cls, data: JsonDict) -> 'Overloaded':
         assert data['.class'] == 'Overloaded'
         return Overloaded([CallableType.deserialize(t) for t in data['items']])
 
@@ -734,11 +770,13 @@ class TupleType(Type):
     implicit = False
 
     def __init__(self, items: List[Type], fallback: Instance, line: int = -1,
-                 implicit: bool = False) -> None:
+                 column: int = -1, implicit: bool = False) -> None:
         self.items = items
         self.fallback = fallback
         self.implicit = implicit
-        super().__init__(line)
+        self.can_be_true = len(self.items) > 0
+        self.can_be_false = len(self.items) == 0
+        super().__init__(line, column)
 
     def length(self) -> int:
         return len(self.items)
@@ -760,6 +798,18 @@ class TupleType(Type):
                          Instance.deserialize(data['fallback']),
                          implicit=data['implicit'])
 
+    def copy_modified(self, *, fallback: Instance = None,
+                  items: List[Type] = None) -> 'TupleType':
+        if fallback is None:
+            fallback = self.fallback
+        if items is None:
+            items = self.items
+        return TupleType(items, fallback, self.line, self.column)
+
+    def slice(self, begin: int, stride: int, end: int) -> 'TupleType':
+        return TupleType(self.items[begin:end:stride], self.fallback,
+                         self.line, self.column, self.implicit)
+
 
 class StarType(Type):
     """The star type *type_parameter.
@@ -769,9 +819,9 @@ class StarType(Type):
 
     type = None  # type: Type
 
-    def __init__(self, type: Type, line: int = -1) -> None:
+    def __init__(self, type: Type, line: int = -1, column: int = -1) -> None:
         self.type = type
-        super().__init__(line)
+        super().__init__(line, column)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_star_type(self)
@@ -782,14 +832,16 @@ class UnionType(Type):
 
     items = None  # type: List[Type]
 
-    def __init__(self, items: List[Type], line: int = -1) -> None:
+    def __init__(self, items: List[Type], line: int = -1, column: int = -1) -> None:
         self.items = items
-        super().__init__(line)
+        self.can_be_true = any(item.can_be_true for item in items)
+        self.can_be_false = any(item.can_be_false for item in items)
+        super().__init__(line, column)
 
     @staticmethod
-    def make_union(items: List[Type], line: int = -1) -> Type:
+    def make_union(items: List[Type], line: int = -1, column: int = -1) -> Type:
         if len(items) > 1:
-            return UnionType(items, line)
+            return UnionType(items, line, column)
         elif len(items) == 1:
             return items[0]
         else:
@@ -799,7 +851,7 @@ class UnionType(Type):
                 return Void()
 
     @staticmethod
-    def make_simplified_union(items: List[Type], line: int = -1) -> Type:
+    def make_simplified_union(items: List[Type], line: int = -1, column: int = -1) -> Type:
         while any(isinstance(typ, UnionType) for typ in items):
             all_items = []  # type: List[Type]
             for typ in items:
@@ -814,13 +866,23 @@ class UnionType(Type):
 
         from mypy.subtypes import is_subtype
         removed = set()  # type: Set[int]
-        for i in range(len(items)):
-            if any(is_subtype(items[i], items[j]) for j in range(len(items))
-                   if j not in removed and j != i):
-                removed.add(i)
+        for i, ti in enumerate(items):
+            if i in removed: continue
+            # Keep track of the truishness info for deleted subtypes which can be relevant
+            cbt = cbf = False
+            for j, tj in enumerate(items):
+                if i != j and is_subtype(tj, ti):
+                    removed.add(j)
+                    cbt = cbt or tj.can_be_true
+                    cbf = cbf or tj.can_be_false
+            # if deleted subtypes had more general truthiness, use that
+            if not ti.can_be_true and cbt:
+                items[i] = true_or_false(ti)
+            elif not ti.can_be_false and cbf:
+                items[i] = true_or_false(ti)
 
         simplified_set = [items[i] for i in range(len(items)) if i not in removed]
-        return UnionType.make_union(simplified_set)
+        return UnionType.make_union(simplified_set, line, column)
 
     def length(self) -> int:
         return len(self.items)
@@ -896,7 +958,7 @@ class EllipsisType(Type):
         return {'.class': 'EllipsisType'}
 
     @classmethod
-    def deserialize(self, data: JsonDict) -> 'EllipsisType':
+    def deserialize(cls, data: JsonDict) -> 'EllipsisType':
         assert data['.class'] == 'EllipsisType'
         return EllipsisType()
 
@@ -933,8 +995,8 @@ class TypeType(Type):
     # a generic class instance, a union, Any, a type variable...
     item = None  # type: Type
 
-    def __init__(self, item: Type, *, line: int = -1) -> None:
-        super().__init__(line)
+    def __init__(self, item: Type, *, line: int = -1, column: int = -1) -> None:
+        super().__init__(line, column)
         self.item = item
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
@@ -1072,7 +1134,7 @@ class TypeTranslator(TypeVisitor[Type]):
         return t
 
     def visit_instance(self, t: Instance) -> Type:
-        return Instance(t.type, self.translate_types(t.args), t.line)
+        return Instance(t.type, self.translate_types(t.args), t.line, t.column)
 
     def visit_type_var(self, t: TypeVarType) -> Type:
         return t
@@ -1088,13 +1150,13 @@ class TypeTranslator(TypeVisitor[Type]):
     def visit_tuple_type(self, t: TupleType) -> Type:
         return TupleType(self.translate_types(t.items),
                          cast(Any, t.fallback.accept(self)),
-                         t.line)
+                         t.line, t.column)
 
     def visit_star_type(self, t: StarType) -> Type:
-        return StarType(t.type.accept(self), t.line)
+        return StarType(t.type.accept(self), t.line, t.column)
 
     def visit_union_type(self, t: UnionType) -> Type:
-        return UnionType(self.translate_types(t.items), t.line)
+        return UnionType(self.translate_types(t.items), t.line, t.column)
 
     def visit_ellipsis_type(self, t: EllipsisType) -> Type:
         return t
@@ -1117,7 +1179,7 @@ class TypeTranslator(TypeVisitor[Type]):
         return Overloaded(items=items)
 
     def visit_type_type(self, t: TypeType) -> Type:
-        return TypeType(t.item.accept(self), line=t.line)
+        return TypeType(t.item.accept(self), line=t.line, column=t.column)
 
 
 class TypeStrVisitor(TypeVisitor[str]):
@@ -1379,15 +1441,112 @@ def strip_type(typ: Type) -> Type:
         return typ
 
 
-def replace_leading_arg_type(t: CallableType, self_type: Type) -> CallableType:
-    """Return a copy of a callable type with a different self argument type.
-
-    Assume that the callable is the signature of a method.
-    """
-    return t.copy_modified(arg_types=[self_type] + t.arg_types[1:])
-
-
 def is_named_instance(t: Type, fullname: str) -> bool:
     return (isinstance(t, Instance) and
             t.type is not None and
             t.type.fullname() == fullname)
+
+
+def copy_type(t: Type) -> Type:
+    """
+    Build a copy of the type; used to mutate the copy with truthiness information
+    """
+    return copy.copy(t)
+
+
+def true_only(t: Type) -> Type:
+    """
+    Restricted version of t with only True-ish values
+    """
+    if not t.can_be_true:
+        # All values of t are False-ish, so there are no true values in it
+        return UninhabitedType(line=t.line, column=t.column)
+    elif not t.can_be_false:
+        # All values of t are already True-ish, so true_only is idempotent in this case
+        return t
+    elif isinstance(t, UnionType):
+        # The true version of a union type is the union of the true versions of its components
+        new_items = [true_only(item) for item in t.items]
+        return UnionType.make_simplified_union(new_items, line=t.line, column=t.column)
+    else:
+        new_t = copy_type(t)
+        new_t.can_be_false = False
+        return new_t
+
+
+def false_only(t: Type) -> Type:
+    """
+    Restricted version of t with only False-ish values
+    """
+    if not t.can_be_false:
+        # All values of t are True-ish, so there are no false values in it
+        return UninhabitedType(line=t.line)
+    elif not t.can_be_true:
+        # All values of t are already False-ish, so false_only is idempotent in this case
+        return t
+    elif isinstance(t, UnionType):
+        # The false version of a union type is the union of the false versions of its components
+        new_items = [false_only(item) for item in t.items]
+        return UnionType.make_simplified_union(new_items, line=t.line, column=t.column)
+    else:
+        new_t = copy_type(t)
+        new_t.can_be_true = False
+        return new_t
+
+
+def true_or_false(t: Type) -> Type:
+    """
+    Unrestricted version of t with both True-ish and False-ish values
+    """
+    new_t = copy_type(t)
+    new_t.can_be_true = type(new_t).can_be_true
+    new_t.can_be_false = type(new_t).can_be_false
+    return new_t
+
+
+def function_type(func: mypy.nodes.FuncBase, fallback: Instance) -> FunctionLike:
+    if func.type:
+        assert isinstance(func.type, FunctionLike)
+        return func.type
+    else:
+        # Implicit type signature with dynamic types.
+        # Overloaded functions always have a signature, so func must be an ordinary function.
+        assert isinstance(func, mypy.nodes.FuncItem), str(func)
+        fdef = cast(mypy.nodes.FuncItem, func)
+        name = func.name()
+        if name:
+            name = '"{}"'.format(name)
+
+        return CallableType(
+            [AnyType()] * len(fdef.arg_names),
+            fdef.arg_kinds,
+            fdef.arg_names,
+            AnyType(),
+            fallback,
+            name,
+            implicit=True,
+        )
+
+
+def get_typ_args(tp: Type) -> List[Type]:
+    """Get all type arguments from a parameterizable Type."""
+    if not isinstance(tp, (Instance, UnionType, TupleType, CallableType)):
+        return []
+    typ_args = (tp.args if isinstance(tp, Instance) else
+                tp.items if not isinstance(tp, CallableType) else
+                tp.arg_types + [tp.ret_type])
+    return typ_args
+
+
+def set_typ_args(tp: Type, new_args: List[Type], line: int = -1, column: int = -1) -> Type:
+    """Return a copy of a parameterizable Type with arguments set to new_args."""
+    if isinstance(tp, Instance):
+        return Instance(tp.type, new_args, line, column)
+    if isinstance(tp, TupleType):
+        return tp.copy_modified(items=new_args)
+    if isinstance(tp, UnionType):
+        return UnionType.make_simplified_union(new_args, line, column)
+    if isinstance(tp, CallableType):
+        return tp.copy_modified(arg_types=new_args[:-1], ret_type=new_args[-1],
+                                line=line, column=column)
+    return tp
